/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * ============================================
 *
 * Copyright (C) 2007-2010 by Tidalwave s.a.s.
 * Project home page: http://openbluesky.kenai.com
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: Aspect.java,v f17245cd751a 2010/03/14 17:14:28 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.aspect;

import javax.annotation.Nonnull;
import java.util.HashSet;
import java.util.Set;
import org.openide.util.Lookup;
import it.tidalwave.util.logging.Logger;

/***********************************************************************************************************************
 *
 * @author Fabrizio Giudici
 * @version $Id: Aspect.java,v f17245cd751a 2010/03/14 17:14:28 fabrizio $
 *
 **********************************************************************************************************************/
public abstract class Aspect
  {
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public static abstract class Task<T, E extends Throwable>
      {
        @Nonnull
        private final String name;

        public Task()
          {
            name = String.format("%s@%x", getClass().getName(), System.identityHashCode(this));
          }

        public Task (final @Nonnull String name)
          {
            this.name = String.format("Aspect.Task@%x[%s]", System.identityHashCode(this), name);
          }

        public abstract T run()
          throws E;

        @Override @Nonnull
        public String toString()
          {
            return name;
          }
      }

    private static final String CLASS = Aspect.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);

    public final static Aspect VOID = new Aspect()
      {
        @Override @Nonnull
        public String toString()
          {
            return "Aspect.VOID";
          }
      };
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public static <T, E extends Throwable> T run (final @Nonnull Lookup.Provider object,
                                                  final @Nonnull Task<T, E> task)
      throws E
      {
        return run(object.getLookup(), task);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public static <T, E extends Throwable> T run (final @Nonnull Lookup lookup,
                                                  final @Nonnull Task<T, E> task)
      throws E
      {
        final Set<Aspect> aspects = new HashSet<Aspect>();
        
        for (final AspectFactory aspectFactory : lookup.lookupAll(AspectFactory.class))
          {
            aspects.add(aspectFactory.createAspect());
          }

        return new MultiAspect(aspects).run(task);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    final public <T, E extends Throwable> T run (final @Nonnull Task<T, E> task)
      throws E
      {
        logger.fine("run(%s) with %s", task, this);

        try
          {
            runBefore();
            final T result = task.run();
            runAfter();
            return result;
          }
        catch (Throwable t)
          {
            logger.throwing(CLASS, "run()", t);
            handleException(t);

            try
              {
                throw (E)t;
              }
            catch (ClassCastException e)
              {
                throw new RuntimeException(e);
              }
          }
        finally
          {
            runFinally();
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void runBefore()
      {
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void runAfter()
      {
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void runFinally()
      {
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected void handleException (final @Nonnull Throwable t)
      {
      }
  }
